/**
	Methods definitions for the
	Knight Chess Problem
	Fraga 05/11/2000
**/
#include "knight.h"
#include <qapplication.h>
#include <qpainter.h>
#include <qdrawutil.h>
#include <qcombobox.h>
#include <qcheckbox.h>
#include <qlabel.h>
#include <qspinbox.h>
#include <qstring.h>
#include <qpixmap.h>


//***************************************************************************
//* ChessButton member functions
//***************************************************************************


ChessButton::ChessButton( QWidget *parent ) : QButton( parent )
{
    t = Blank;	// initial type
	val = 0;
	setFocusPolicy(NoFocus); // We don't use the keyboard on this Botton
}

// --------------------------------------------------------------------------
// Paints ChessButton
//
void ChessButton::drawButton( QPainter *p )
{
	QRect r = rect();				// get rectangle
	static QColorGroup g(white, blue, white, darkBlue, blue, black, black );

	QBrush fill;

	fill.setStyle( SolidPattern );

 	if ( c == White )  fill.setColor( yellow );
 	else fill.setColor( green );

	qDrawShadePanel( p, r, g, isDown(), 1, &fill );
//	qDrawPlainRect ( p, r, black, 1, &fill );

	if ( t == Passed ) {   // Draw val value on the botton
		p->setPen( Qt::black );
		p->setFont( QFont( "times", 18, QFont::Bold ) );
		QString s;
		// p->drawText( r, AlignCenter, s.arg( val ) );
		p->drawText( r, AlignCenter, s.sprintf("%d", val ) );
	}
	else if( t == Actual ){ 

		p->drawPixmap( 1, 1, (QPixmap &)* img );
	}	
}


//***************************************************************************
//* ChessGameBoard member functions
//***************************************************************************

// --------------------------------------------------------------------------
// Creates a game board with N x N buttons and connects the "clicked()"
// signal of all buttons to the "buttonClicked()" slot.
//

ChessGameBoard::ChessGameBoard( int n, QPixmap *mypixmap,
	QWidget *parent, const char *name ) : QWidget( parent, name )
{
	aa = (int[]){ 2, 1,-1,-2,-2,-1, 1, 2}; /** Horizontal **/
	bb = (int[]){-1,-2,-2,-1, 1, 2, 2, 1}; /** Vertical   **/
	
    st = Init;					// initial state
    nBoard = n;
    n *= n;					// make square
    comp_starts = FALSE;			// Computer can starts
    buttons = new ChessButtons(n);		// create real buttons

	calculate_accessboard( );

    for ( int i=0; i<n; i++ ) {			// create and connect buttons
		ChessButton *p = new ChessButton( this );
		connect( p, SIGNAL(clicked()), SLOT(buttonClicked()) );

		int row = i/nBoard;
		int col = i%nBoard;
		int color = (row%2) ?
			(col%2) ? 1 : 0
			: (col%2) ? 0 : 1;
		if ( color ) p->setColor( ChessButton::White );
		else  p->setColor( ChessButton::Black );

 		if ( color ) p->setBackgroundColor( yellow );
 		else p->setBackgroundColor( green );

		int x = 60*col; // This coordinates are respect to father widget
		int y = 60*row;

		p->setGeometry( x, y, 60, 60 ); 
	
		// p->setValue( i ); // It was a little test :-)

		p->img = mypixmap;

		buttons->insert( i, p );
    }

    setIniPos( 1, 1 );
    setCounter( 1 );
	int iold = nBoard * (nBoard - 1) ;
	ChessButton *a = buttons->at(iold);   // get piece at that row & col
	a->setValue( 1 );
	a->setType( ChessButton::Actual );
	playing = FALSE;

	setFocusPolicy(NoFocus);
}

ChessGameBoard::~ChessGameBoard()
{
    delete buttons;
}


// --------------------------------------------------------------------------
// ChessGameBoard::computerStarts( bool v )
//
// Computer starts if v=TRUE. The human starts by default.
//

void ChessGameBoard::computerStarts( bool v )
{
    comp_starts = v;
}


// --------------------------------------------------------------------------
// ChessGameBoard::newGame()
//
// Clears the game board and prepares for a new game
//

void ChessGameBoard::newGame()
{
	if ( st == Playing ){
		playing = FALSE;
		killTimers( );
	}
	st = Init;

	calculate_accessboard( );

	for ( int i=0; i<nBoard*nBoard; i++ )
		buttons->at(i)->setType( ChessButton::Blank );

	int i, j;
	for( i=0; i<nBoard; i++ ) 
		for( j=0; j<nBoard; j++ ) h[i][j] = 0;

	buttons->at(nBoard*(nBoard-1))->setType( ChessButton::Actual );

	setNewPos( 1, 1 );
	setCounter( 1 );
	emit newPosition( );

	// If the new position was the same that the initial:
	setIniPos( 1, 1 );
	emit newMessagePosition( );
}

void ChessGameBoard::calculate_accessboard( )
{
	int i, j, k, l, m, c;
	int u, v;

	l = nBoard-1; 
	k = nBoard/2;

	for( i=0; i<k; i++ ) {
		for( j=0; j<k; j++ ) {
			for ( c=0, m=0; m<8; m++ ) {
				u=j+aa[m]; /** horizontal movement **/  
				v=i+bb[m]; /** vertical  movement **/
				if(0<=u && u<nBoard && 0<=v && v<nBoard ) c++; 
			}
			accessboard[ i ][ j ] = c;
			accessboard[ i ][l-j] = c;
			accessboard[l-i][ j ] = c;
			accessboard[l-i][l-j] = c;
		}
	}
	if ( nBoard%2 ) {  // The board has odd size
		for( i=0; i<=k; i++ ) {
			for ( c=0, m=0; m<8; m++ ) {
				u=i+aa[m]; /** horizontal movement **/  
				v=k+bb[m]; /** vertical  movement **/
				if(0<=u && u<nBoard && 0<=v && v<nBoard ) c++; 
			}
			accessboard[ i ][ k ] = c;
			accessboard[ k ][ i ] = c;
			accessboard[ k ][l-i] = c;
			accessboard[l-i][ k ] = c;
		}
	}
//	fprintf( stderr, "\n" ); 
//	for( i=0; i<=l; i++ ) {
//		for( j=0; j<=l; j++ ) fprintf( stderr, "%d ", accessboard[i][j] ); 
//		fprintf( stderr, "\n" ); 
//	}
}

// --------------------------------------------------------------------------
// ChessGameBoard::buttonClicked()		- SLOT
//
// This slot is activated when a ChessButton emits the signal "clicked()",
// i.e. the user has clicked on a ChessButton.
//

void ChessGameBoard::buttonClicked()
{
	int inew, row, col;

    if ( st == Init ){ // Draw (choose) initial potition	
		
		// New pressed button 
		inew = buttons->findRef( (ChessButton *)sender() );

		row = nBoard - inew/nBoard;
		col = inew%nBoard + 1;

		setNewPos( row, col );

		emit newPosition( );
	}
	else if ( st == Playing || st == Badmove ) {  // The human is playing 
		if ( st == Badmove ) st = Playing;
		emit finished( );

		// New pressed button 
		inew = buttons->findRef( (ChessButton *)sender() );
		row = inew/nBoard;
		col = inew%nBoard;

		TryMove( row, col ); 

	}
}

void  ChessGameBoard::TryMove( int row, int col )
{
	int inew = nBoard*row + col;

	if ( test_movement( row, col ) == TRUE ){
		// Now we check if we can to make the next trial 

		bool halted = check_next_trial(  row, col );

		h[row][col] = 1;

		// The last button pressed
		int r = nBoard - irow; 
		int c = icol - 1;
		ChessButton *a = buttons->at(nBoard*r + c); 
	   	a->setValue( counter );
		a->setType( ChessButton::Passed );

		counter ++;
		row = nBoard - inew/nBoard;
		col = inew%nBoard + 1;

		setIniPos( row, col );
			
		ChessButton *b = buttons->at(inew); // the new button pressed
    	b->setValue( counter );
		b->setType( ChessButton::Actual );

		emit newMessagePosition( );

		if ( halted == TRUE  ) { // The game is halted
			st = Gameover;
			emit finished( );
 		}
	}
	else {
		st = Badmove;
		emit finished( );
	}
}

bool ChessGameBoard::test_movement( int row, int col ) 
{
	bool success = FALSE;

	int u, v, i;

	int r = nBoard - irow; 
	int c = icol - 1;
		
	for (  i=0; i<8; i++ ){
		u = c + aa[i];  /** Horizontal movement **/	
		v = r + bb[i];  /** Vertical movement **/

		if ( 0<=u && u<nBoard && 0<=v && v<nBoard ) {

			if ( row==v && col==u && h[v][u]==0 ) {
				success = TRUE;
				break;
			}
		}
	}
	return( success );
}

bool ChessGameBoard::check_next_trial( int row, int col )
{
	int u, v, i;
	int busy = 0;
	int possiblemoves=0;

	for (  i=0; i<8; i++ ){
		u = col + aa[i];  /** Horizontal movement **/	
		v = row + bb[i];  /** Vertical movement **/
		if ( 0<=u && u<nBoard && 0<=v && v<nBoard ){
			possiblemoves++; 
			if ( h[v][u]==1 ) busy++;
		}
	}
	if ( busy == possiblemoves ) return TRUE;
	else return FALSE;
}

void ChessGameBoard::ChangeButtons( int newrow, int newcol )
{
	// Get actual coordinates
	int row = nBoard - getPosRow( );
	int col = getPosCol( ) - 1;

	int iold = nBoard * row + col;
	ChessButton *a = buttons->at(iold);   // get piece that was pressed
    a->setValue( 0 );
    a->setType( ChessButton::Blank );

	// Get new coodinates
	int inew = nBoard * (nBoard - newrow) + newcol - 1;
	ChessButton *b = buttons->at(inew);   // get piece that was pressed
    b->setValue( 1 );
    b->setType( ChessButton::Actual );

	setIniPos( newrow, newcol );
}


void ChessGameBoard::PlayGame( )
{
	if ( st != Init ) return;

	st = Playing;
	emit finished( );


	if ( comp_starts )
		computerPlay( );
	else {
		int r = nBoard - getPosRow( );
		int c = getPosCol( ) - 1;
		h[r][c] = 1;	
	}
}



// --------------------------------------------------------------------------
// ChessGameBoard::computerPlay()
//
// Init timer for computer's pieces move
//
void ChessGameBoard::computerPlay()
{
	if ( playing ) return;

	playing = TRUE;
	startTimer(1200);
}

void ChessGameBoard::timerEvent( QTimerEvent * )
{
	int r, c, i;
	int wherer, wherec;
	int succes;

	// initial position
	r = nBoard - getPosRow( );
	c = getPosCol( ) - 1;
	h[r][c] = 1;
	i = nBoard*r + c;  
	ChessButton *a = buttons->at(i);   // get piece that was pressed
    a->setValue( counter );
	if ( counter < nBoard*nBoard )        // Leave the last move untouched
		a->setType( ChessButton::Passed );

	succes = trymove( r,c, &wherer, &wherec);

	if ( succes == FALSE || counter>=nBoard*nBoard ) {
		playing = FALSE;
		killTimers( );

		st = Gameover;
		emit finished( );

		return;
	}

	counter ++;
	int row = nBoard - wherer;
	int col = wherec + 1; 
	setIniPos( row, col );

	i = nBoard*wherer + wherec;  

	ChessButton *b = buttons->at(i);   // get piece that was pressed
    b->setValue( counter );
	b->setType( ChessButton::Actual );
	emit newMessagePosition( );
}


int ChessGameBoard::trymove( int y, int x, int *to_y, int *to_x)
{
	int i, j, k, u, v;

	int l = nBoard;

	/** Vemos a donde tirar, contruyendo primero
       la "accesibilidad" **/
	for ( i=0; i<8; i++ ) {
		u=x+aa[i]; /** horizontal movement **/  
		v=y+bb[i]; /** vertical  movement **/
		if(0<=u && u<l && 0<=v && v<l && h[v][u]==0){
			accessibility[i] = accessboard[v][u];
		}
		else {
			accessibility[i] = 10;
		}
	}

	/** Search the minor "accesibility" **/
	k=100;
	for ( i=0; i<8; i++ ) {
		if ( accessibility[i] < k ) {
			k = accessibility[i];
			j = i;
		}
	}

	if ( accessibility[j] < 10 ) {
		u=x+aa[j]; /** horizontal movement **/  
		v=y+bb[j]; /** vertical movement  **/

		h[v][u]=1;

		*to_y = v;
		*to_x = u;

		for ( i=0; i<8; i++ ) {
			u=x+aa[i]; /** horizontal movement **/  
			v=y+bb[i]; /** vertical movement   **/

			if(0<=u && u<l && 0<=v && v<l ){
				accessboard[v][u] -= 1;
			}
		}
		return( TRUE );
	}
	else return( FALSE );
}    




//***************************************************************************
//* ChessGame member functions
//***************************************************************************

// --------------------------------------------------------------------------
// Creates a game widget with a game board and two push buttons, and connects
// signals of child widgets to slots.
//

ChessGame::ChessGame( int boardSize, QPixmap *pixmap,
		QWidget *parent, const char *name ) : QWidget( parent, name )
{
	int n = 60*boardSize;
	int vertical = ( n < 280) ? 280 : n ;
	setFixedSize( n+130, vertical ); // We can't resize

// Create the game board and connect the signal finished() to this
// gameOver() slot

// create and connect widgets
    board= new ChessGameBoard(boardSize,pixmap,this);	
	board->setGeometry( 130, 0, n, n );	// resize the chess game board
    connect( board, SIGNAL(finished()), SLOT(gameOver()) );
    connect( board, SIGNAL(newPosition()), SLOT(changePosition()) );
    connect( board, SIGNAL(newMessagePosition()), SLOT(setMessPoss()) );

// A vertical line between the buttons and the chess board
    QFrame *linev = new QFrame( this );
    linev->setGeometry( 120, 5, 10, vertical-10 );
    linev->setFrameStyle( QFrame::VLine | QFrame::Sunken );

// Create a message label

    message = new QLabel( this );
    message->setGeometry( 10, 10, 105, 20 );
    message->setFrameStyle( QFrame::WinPanel | QFrame::Sunken );
    message->setBackgroundColor( message->colorGroup().base() );
    message->setAlignment( AlignCenter );

// Create the push buttons and connect their clicked() signals
// to this buttonClicked() slot

    startGame = new QPushButton( "Start", this );
    startGame->setGeometry( 30, 40, 70, 25 );
    connect( startGame, SIGNAL(clicked()), SLOT(startGameClicked()) );

// Create the combo box for deciding who should start, and
// connect its clicked() signals to the buttonClicked() slot
    
    whoPlay = new QComboBox( this );
    whoPlay->insertItem( "Computer play" );
    whoPlay->insertItem( "Human play" );
    whoPlay->move( 0,0 );
    whoPlay->adjustSize();
    whoPlay->move( 5, 75 );

// Create a horizontal frame line

    QFrame *lineh1 = new QFrame( this );
    lineh1->setGeometry( 5, 105, 115, 10 );
    lineh1->setFrameStyle( QFrame::HLine | QFrame::Sunken );

// Create two QSpinBoxes to put the initial coordinates

	sbox1 = new QSpinBox( 1, boardSize, 1, this, "sbox1" );
	sbox1->setGeometry( 15, 125, 45, 25 );
	sbox1->setWrapping( TRUE );
	connect( sbox1, SIGNAL(valueChanged(int)), SLOT(spin1Clicked( )) );

	sbox2 = new QSpinBox( 1, boardSize, 1, this, "sbox2" );
	sbox2->setGeometry( 70, 125, 45, 25 );
	sbox2->setWrapping( TRUE );
	connect( sbox2, SIGNAL(valueChanged(int)), SLOT(spin2Clicked( )) );

// The position is displayed

    messposs = new QLabel( this );
    messposs->setGeometry( 15, 160, 100, 20 );
    messposs->setFrameStyle( QFrame::WinPanel | QFrame::Sunken );
    messposs->setBackgroundColor( message->colorGroup().base() );
    messposs->setAlignment( AlignCenter );
//
    QFrame *lineh2 = new QFrame( this );
    lineh2->setGeometry( 5, 190, 115, 10 );
    lineh2->setFrameStyle( QFrame::HLine | QFrame::Sunken );

// Create the push buttons and connect their clicked() signals
// to this buttonClicked() slot

    newGame = new QPushButton( "NewGame", this );
    newGame->setGeometry( 30, 210, 70, 25 );
    connect( newGame, SIGNAL(clicked()), SLOT(newGameClicked()) );
//
    quit = new QPushButton( "Quit", this );
    quit->setGeometry( 30, 245, 70, 25 );
    connect( quit, SIGNAL(clicked()), qApp, SLOT(quit()) );

	setMessPoss ( );
	// Start the game
    newState();

	flagsboxes = 0;
}


// --------------------------------------------------------------------------
// ChessGame::newGameClicked()			- SLOT
//
// This slot is activated when the new game button is clicked.
//

void ChessGame::newGameClicked()
{
    board->newGame( );
    newState();
}

void ChessGame::startGameClicked()
{
    board->computerStarts( whoPlay->currentItem() == 0 );
    board->PlayGame();
}

void ChessGame::gameOver()    // -SLOT
{
    newState();
}

void ChessGame::changePosition( ) // -SLOT
{
	int orow, ocol, ncol, nrow;

	orow = board->getPosRow( );
	ocol = board->getPosCol( );

	board->getNewPos( &nrow, &ncol );

	if ( nrow != orow ) 
		sbox2->setValue( nrow );
	if ( ncol != ocol ) 
		sbox1->setValue( ncol );
	
}

void ChessGame::setMessPoss( ) // - SLOT in initial message
{
	static char acol[8] = {'a','b','c','d','e','f','g','h' };
	int row = board->getPosRow( );
	int col = board->getPosCol( );
	
	QString s;
	messposs->setText( s.sprintf("%c%d", acol[col-1], row ) );
}


// ChessGame::spin1Clicked()			- SLOT
//
// This slot is activated when the spin box 1  is clicked.
//
void ChessGame::spin1Clicked()
{
	int row, newcol;
	row = board->getPosRow( );
	newcol = sbox1->value( );

	if ( board->state() == ChessGameBoard::Init ) {
		board->ChangeButtons( row, newcol );
		setMessPoss( );
	}
	HumanPlay( row, newcol );
}

// ChessGame::spin2Clicked()			- SLOT
//
// This slot is activated when the spin box 2  is clicked.

void ChessGame::spin2Clicked()
{
	int newrow = sbox2->value( );
	int col = board->getPosCol( );

	if ( board->state() == ChessGameBoard::Init ) {
		board->ChangeButtons( newrow, col );
		setMessPoss( );
	}
	HumanPlay( newrow, col );
}

void ChessGame::HumanPlay( int row, int col ) {
	if ( board->state() == ChessGameBoard::Playing ||
	     board->state() == ChessGameBoard::Badmove ) {
		if ( !board -> WhoPlay( ) ) {
			if ( !flagsboxes ) { 
				flagsboxes++;
				board->ChangeButtons( row, col );
				setMessPoss( );
			}
			else {
				board->TryMove( row, col );
				flagsboxes = 0;
			}
		}
	}
}

// Updates the message to reflect a new state.
//

void ChessGame::newState()
{
    static char *msg[] = {			// ChessGameBoard::State texts
	"Wanna play?", "Playing", "Game Over", "Bad Move" };
    message->setText( msg[board->state()] );
    return;
}
